/*
** File dialog functions
**
** Copyright  1997 Marko Mkel, see COPYING.LIB for details
*/

#include <stdio.h>
#include <stdlib.h>

#include <sys/stat.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/AsciiText.h>

#include "util.h"
#include "filedialog.h"

#include "FChooser.h"

#define ActionProc(name)\
static void name (Widget w, XEvent *event, String *params, Cardinal *nparams)

ActionProc(okAction);

#define CallbackFunc(name)\
static  void name (Widget w, XtPointer closure, XtPointer call_data)

CallbackFunc(okCallback);
CallbackFunc(cancelCallback);
CallbackFunc(selectionChangedCallback);
CallbackFunc(sortModeCallback);

static Widget openFileShell = NULL;
static Widget chooser;
static Widget fileText;

/* Variables for reading the start and end addresses */
static Widget startAddr = NULL;
static Widget endAddr = NULL;
char StartAddrString [5], EndAddrString [5];
unsigned StartAddr, EndAddr;

static XtActionsRec actionTable[] = {
  { "ok",	okAction },
};

/* File name returned by openFile() and writeFile() */
static char *openFileName = NULL;

/* Checks to perform on the file name in okCallback() */
enum { DontCare, MustExist, MustNotExist } SeeIfExists;

/* Ask the name of a file to be read */
char *openFile (Widget parent)
{
  Widget form, button, menu, menuItem, label;
  int i;
  Arg args[3];

  XtAppContext appContext = XtWidgetToApplicationContext (parent);

  if (openFileShell || openFileName)
    return NULL; /* this is a non-reentrant function */

  SeeIfExists = MustExist;

  XtAppAddActions (appContext, actionTable, XtNumber(actionTable));

  XtSetArg (args[0], XtNinput, True);

  openFileShell = XtCreatePopupShell ("openFile", transientShellWidgetClass,
				      parent, args, 1);

  form = XtCreateManagedWidget ("form", formWidgetClass,
				openFileShell, NULL, 0);

  chooser = XtCreateManagedWidget ("chooser", xfwfFileChooserWidgetClass,
				   form, NULL, 0);
  XtAddCallback (chooser, XtNcallback, selectionChangedCallback, NULL);

  XtSetArg (args[0], XtNfromHoriz, chooser);
  button = XtCreateManagedWidget ("okButton", commandWidgetClass,
				  form, args, 1);
  XtAddCallback (button, XtNcallback, okCallback, NULL);

  XtSetArg (args[0], XtNfromHoriz, chooser);
  XtSetArg (args[1], XtNfromVert, button);
  button = XtCreateManagedWidget ("cancelButton", commandWidgetClass,
				  form, args, 2);
  XtAddCallback (button, XtNcallback, cancelCallback, NULL);

  menu = XtCreatePopupShell ("sortMenu", simpleMenuWidgetClass,
			     form, NULL, 0);

  for (i = 1; i < 7; i++) {
    char itemname[15];
    sprintf (itemname, "sortMenuItem%u", i);

    menuItem = XtCreateManagedWidget (itemname, smeBSBObjectClass,
				      menu, NULL, 0);
    XtAddCallback (menuItem, XtNcallback, sortModeCallback, (XtPointer)i);
  }

  XtSetArg (args[0], XtNfromHoriz, chooser);
  XtSetArg (args[1], XtNfromVert, button);
  XtSetArg (args[2], XtNmenuName, "sortMenu");
  button = XtCreateManagedWidget ("sort", menuButtonWidgetClass,
				  form, args, 3);

  XtSetArg (args[0], XtNfromVert, chooser);
  label = XtCreateManagedWidget ("fileLabel", labelWidgetClass,
				 form, args, 1);

  XtSetArg (args[0], XtNfromVert, chooser);
  XtSetArg (args[1], XtNfromHoriz, label);
  XtSetArg (args[2], XtNeditType, XawtextEdit);
  fileText = XtCreateManagedWidget ("fileText", asciiTextWidgetClass,
				    form, args, 3);

  Popup (openFileShell, parent);

  while (XtIsRealized (openFileShell))
    XtAppProcessEvent (appContext, XtIMAll);

  XtDestroyWidget (openFileShell);
  openFileShell = NULL;

  {
    char *filename = openFileName;
    openFileName = NULL;
    return filename;
  }
}

/* Ask the user for the name of a file to be written.
   Optionally ask the file's start and end addresses. */
char *writeFile (Widget parent, Boolean mayExist,
		 unsigned *start, unsigned *end)
{
  Widget form, button, menu, menuItem, label;
  int i;
  Arg args[3];

  XtAppContext appContext = XtWidgetToApplicationContext (parent);

  if (openFileShell || openFileName)
    return NULL; /* this is a non-reentrant function */

  SeeIfExists = mayExist ? DontCare : MustNotExist;

  XtAppAddActions (appContext, actionTable, XtNumber(actionTable));

  XtSetArg (args[0], XtNinput, True);

  openFileShell = XtCreatePopupShell ("writeFile", transientShellWidgetClass,
				      parent, args, 1);

  form = XtCreateManagedWidget ("form", formWidgetClass,
				openFileShell, NULL, 0);

  chooser = XtCreateManagedWidget ("chooser", xfwfFileChooserWidgetClass,
				   form, NULL, 0);
  XtAddCallback (chooser, XtNcallback, selectionChangedCallback, NULL);

  XtSetArg (args[0], XtNfromHoriz, chooser);
  button = XtCreateManagedWidget ("okButton", commandWidgetClass,
				  form, args, 1);
  XtAddCallback (button, XtNcallback, okCallback, NULL);

  XtSetArg (args[0], XtNfromHoriz, chooser);
  XtSetArg (args[1], XtNfromVert, button);
  button = XtCreateManagedWidget ("cancelButton", commandWidgetClass,
				  form, args, 2);
  XtAddCallback (button, XtNcallback, cancelCallback, NULL);

  menu = XtCreatePopupShell ("sortMenu", simpleMenuWidgetClass,
			     form, NULL, 0);

  for (i = 1; i < 7; i++) {
    char itemname[15];
    sprintf (itemname, "sortMenuItem%u", i);

    menuItem = XtCreateManagedWidget (itemname, smeBSBObjectClass,
				      menu, NULL, 0);
    XtAddCallback (menuItem, XtNcallback, sortModeCallback, (XtPointer)i);
  }

  XtSetArg (args[0], XtNfromHoriz, chooser);
  XtSetArg (args[1], XtNfromVert, button);
  XtSetArg (args[2], XtNmenuName, "sortMenu");
  button = XtCreateManagedWidget ("sort", menuButtonWidgetClass,
				  form, args, 3);

  XtSetArg (args[0], XtNfromVert, chooser);
  label = XtCreateManagedWidget ("fileLabel", labelWidgetClass,
				 form, args, 1);

  XtSetArg (args[0], XtNfromVert, chooser);
  XtSetArg (args[1], XtNfromHoriz, label);
  XtSetArg (args[2], XtNeditType, XawtextEdit);
  fileText = XtCreateManagedWidget ("fileText", asciiTextWidgetClass,
				    form, args, 3);

  {
    Widget vertref = button;

    if (start) {
      sprintf (StartAddrString, "%04X", *start);

      label = XtVaCreateManagedWidget ("startAddrLabel", labelWidgetClass,
				       form,
				       XtNfromHoriz, chooser,
				       XtNfromVert, vertref,
				       NULL);

      startAddr = XtVaCreateManagedWidget ("startAddr", asciiTextWidgetClass,
					   form,
					   XtNfromHoriz, label,
					   XtNfromVert, vertref,
					   XtNuseStringInPlace, True,
					   XtNstring, StartAddrString,
					   XtNlength, sizeof StartAddrString - 1,
					   XtNeditType, XawtextEdit,
					   NULL);

      vertref = label;
    }
    if (end) {
      sprintf (EndAddrString, "%04X", *end);

      label = XtVaCreateManagedWidget ("endAddrLabel", labelWidgetClass,
				       form,
				       XtNfromHoriz, chooser,
				       XtNfromVert, vertref,
				       NULL);

      endAddr = XtVaCreateManagedWidget ("endAddr", asciiTextWidgetClass,
					 form,
					 XtNfromHoriz, label,
					 XtNfromVert, vertref,
					 XtNuseStringInPlace, True,
					 XtNstring, EndAddrString,
					 XtNlength, sizeof EndAddrString - 1,
					 XtNeditType, XawtextEdit,
					 NULL);
    }
  }

  Popup (openFileShell, parent);

  while (XtIsRealized (openFileShell))
    XtAppProcessEvent (appContext, XtIMAll);

  startAddr = endAddr = NULL;

  XtDestroyWidget (openFileShell);
  openFileShell = NULL;

  {
    char *filename = openFileName;
    openFileName = NULL;
    if (start) *start = StartAddr;
    if (end) *end = EndAddr;
    return filename;
  }
}

/*
** Callbacks
*/

CallbackFunc(selectionChangedCallback)
{
  Arg arg;
  XfwfFileChooserReturnStruct *ret =
    (XfwfFileChooserReturnStruct *)call_data;

  if (ret->directory && ret->file)
    XtSetArg (arg, XtNstring, ret->file);
  else
    XtSetArg (arg, XtNstring, "");

  XtSetValues (fileText, &arg, 1);
}

CallbackFunc(okCallback)
{
  Arg arg;
  char *dir, *file;
  struct stat statbuf;

  /* Check the start and end addresses */
  if (startAddr) {
    char *endptr;
    StartAddr = strtoul (StartAddrString, &endptr, 16);
    if (!*StartAddrString || *endptr) {
      XBell (XtDisplayOfObject (openFileShell), 0);
      return;
    }
  }
  if (endAddr) {
    char *endptr;
    EndAddr = strtoul (EndAddrString, &endptr, 16);
    if (!*EndAddrString || *endptr) {
      XBell (XtDisplayOfObject (openFileShell), 0);
      return;
    }
  }

  /* Get the directory from the chooser */
  XtSetArg (arg, XtNcurrentDirectory, &dir);
  XtGetValues (chooser, &arg, 1);

  /* But get file from text item (might differ from XtNcurrentFile) */
  XtSetArg (arg, XtNstring, &file);
  XtGetValues (fileText, &arg, 1);

  /* Check that the file can be read. */
  if (dir && *dir && !stat (dir, &statbuf) && S_ISDIR(statbuf.st_mode) &&
      file && *file) {
    if ((openFileName = malloc (strlen (dir) + strlen (file) + 1))) {
      strcpy (openFileName, dir);
      strcat (openFileName, file);

      if (stat (openFileName, &statbuf)) {
	if (SeeIfExists == DontCare || SeeIfExists == MustNotExist)
	  goto acceptFile;
      }
      else if (S_ISREG (statbuf.st_mode) && SeeIfExists != MustNotExist) {
      acceptFile:
	chdir (dir);
	XtUnrealizeWidget (openFileShell);
	return;
      }

      free (openFileName);
      openFileName = NULL;
    }
  }

  XBell (XtDisplayOfObject (openFileShell), 0);
}

CallbackFunc(cancelCallback)
{
  XtUnrealizeWidget (openFileShell);
}

CallbackFunc(sortModeCallback)
{
  Arg arg;

  XtSetArg (arg, XtNsortMode, (int)closure);
  XtSetValues (chooser, &arg, 1);
}

/*
** Actions
*/

ActionProc(okAction)
{
  okCallback (NULL, NULL, NULL);
}
