/*
** Initialize the X11 user interface
**
** Copyright  1997-1998 Marko Mkel, see COPYING.LIB for details
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <limits.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>

#ifndef NO_SHM
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif /* NO_SHM */

#include "init.h"

#include "util.h"
#include "menu.h"
#include "dialog.h"

#include "keyboard.h"

/* The top level widget */
Widget TopLevel = NULL;

/* Application resources */
AppResources appResources;

/* Flag: has the application been closed? */
Boolean closed = False;

#define BIND(field,class,type,default) \
{ #field, #class, type, sizeof appResources.field, \
  XtOffsetOf(AppResources, field), XtRString, default }
#define BINDARR(field,index,class,type,default) \
{ #field ## #index, #class, type, sizeof appResources.field, \
  XtOffsetOf(AppResources, field[index]), XtRString, default }

static XtResource resources[] = {
  BIND(pausedString,	PausedString,	XtRString, ""),
  BIND(commandString,	CommandString,	XtRString, ""),
  BIND(romPath,		Path,		XtRString, "%s"),
  BIND(filePath,	Path,		XtRString, "%s"),
  BIND(kernalRom,	KernalRom,	XtRString, "kernal.rom"),
  BIND(basicRom,	BasicRom,	XtRString, "basic.rom"),
  BIND(characterRom,	CharacterRom,	XtRString, "char.rom"),
  BIND(exRom,		ExRom,		XtRString, ""),
  BIND(driveRom,	DriveRom,	XtRString, "c1541.rom"),
  BINDARR(color, 0,	Colors,		XtRString, ""),
  BINDARR(color, 1,	Colors,		XtRString, ""),
  BINDARR(color, 2,	Colors,		XtRString, ""),
  BINDARR(color, 3,	Colors,		XtRString, ""),
  BINDARR(color, 4,	Colors,		XtRString, ""),
  BINDARR(color, 5,	Colors,		XtRString, ""),
  BINDARR(color, 6,	Colors,		XtRString, ""),
  BINDARR(color, 7,	Colors,		XtRString, ""),
  BINDARR(color, 8,	Colors,		XtRString, ""),
  BINDARR(color, 9,	Colors,		XtRString, ""),
  BINDARR(color, 10,	Colors,		XtRString, ""),
  BINDARR(color, 11,	Colors,		XtRString, ""),
  BINDARR(color, 12,	Colors,		XtRString, ""),
  BINDARR(color, 13,	Colors,		XtRString, ""),
  BINDARR(color, 14,	Colors,		XtRString, ""),
  BINDARR(color, 15,	Colors,		XtRString, ""),

  BIND(keyRestore,	KeyRestore,	XtRString, ""),
  BIND(keyShiftL,	KeyShiftL,	XtRString, ""),
  BIND(keyShiftR,	KeyShiftR,	XtRString, ""),

  BIND(joy1North,	JoyNorth,	XtRString, ""),
  BIND(joy1Northeast,	JoyNortheast,	XtRString, ""),
  BIND(joy1East,	JoyEast,	XtRString, ""),
  BIND(joy1Southeast,	JoySoutheast,	XtRString, ""),
  BIND(joy1South,	JoySouth,	XtRString, ""),
  BIND(joy1Southwest,	JoySouthwest,	XtRString, ""),
  BIND(joy1West,	JoyWest,	XtRString, ""),
  BIND(joy1Northwest,	JoyNorthwest,	XtRString, ""),
  BIND(joy1Fire,	JoyFire,	XtRString, ""),

  BIND(joy2North,	JoyNorth,	XtRString, ""),
  BIND(joy2Northeast,	JoyNortheast,	XtRString, ""),
  BIND(joy2East,	JoyEast,	XtRString, ""),
  BIND(joy2Southeast,	JoySoutheast,	XtRString, ""),
  BIND(joy2South,	JoySouth,	XtRString, ""),
  BIND(joy2Southwest,	JoySouthwest,	XtRString, ""),
  BIND(joy2West,	JoyWest,	XtRString, ""),
  BIND(joy2Northwest,	JoyNorthwest,	XtRString, ""),
  BIND(joy2Fire,	JoyFire,	XtRString, ""),

  BIND(privateColormap,	PrivateColormap,XtRBoolean, XtEfalse),
  BIND(mitShm,		Output,		XtRBoolean, XtEtrue),
  BIND(cache,		Output,		XtRBoolean, XtEfalse),
  BIND(xsync,		Output,		XtRBoolean, XtEfalse),
  BIND(fixTop,	       	CenterFix,	XtRBoolean, XtEtrue),
  BIND(fixLeft,		CenterFix,	XtRBoolean, XtEtrue),
  BIND(paused,		Paused,		XtRBoolean, XtEfalse),
  BIND(emulate1541,	Emulation,	XtRBoolean, XtEfalse),
  BIND(exrom,		Exrom,		XtRBoolean, XtEfalse),
  BIND(joy1,		Joystick,	XtRBoolean, XtEtrue),
  BIND(joy2,		Joystick,	XtRBoolean, XtEtrue),
  BIND(confirmExit,	Confirm,	XtRBoolean, XtEtrue),
  BIND(topLine,		TopLine,	XtRInt, "0"),
  BIND(bottomLine,	BottomLine,	XtRInt, "0"),
  BIND(leftColumn,	LeftColumn,	XtRInt, "0"),
  BIND(rightColumn,	RightColumn,	XtRInt, "0"),
  BIND(fixTopLine,	TopLine,	XtRInt, "0"),
  BIND(fixLeftColumn,	LeftColumn,	XtRInt, "0"),
  BIND(speed,		Speed,		XtRInt, "100"),
  BIND(driveLed,	Leds,		XtRInt, "0"),
  BIND(horizBlock,	HorizBlock,	XtRInt, "32"),
  BIND(vertBlock,	VertBlock,	XtRInt, "78"),
  BIND(frameParts,	Frames,		XtRInt, "1"),
  BIND(frameSkip,	Frames,		XtRInt, "0"),

  BIND(visual,		Visual,		XtRVisual, NULL),
  BIND(depth,		Depth,		XtRInt, NULL),
};
#undef BIND
#undef BINDARR

/* Command line options for setting some application resources */
#define OPTBOOLEAN(option,resource)				\
{ "-" #option, "." #resource, XrmoptionNoArg, XtEfalse },	\
{ "+" #option, "." #resource, XrmoptionNoArg, XtEtrue }
#define OPT(option,resource)				\
{ "-" #option, "." #resource, XrmoptionSepArg, NULL }

static XrmOptionDescRec optionList[] = {
  OPTBOOLEAN(install, privateColormap),
  OPTBOOLEAN(shm, mitShm),
  OPTBOOLEAN(cache, cache),
  OPTBOOLEAN(xsync, xsync),
  OPTBOOLEAN(paused, paused),
  OPTBOOLEAN(1541, emulate1541),
  OPTBOOLEAN(exrom, exrom),
  OPTBOOLEAN(joy1, joy1),
  OPTBOOLEAN(joy2, joy2),
  OPTBOOLEAN(exit, confirmExit),
  OPT(path, Path),
  OPT(kernal, kernalRom),
  OPT(basic, basicRom),
  OPT(chars, characterRom),
  OPT(cart, exRom),
  OPT(drive, driveRom),
  OPT(speed, speed),
  OPT(parts, frameParts),
  OPT(skip, frameSkip),
  OPT(command, commandString),
  OPT(visual, visual),
  OPT(depth, depth),
};

#undef OPTBOOLEAN
#undef OPT

/* Boolean options for Toggle(), Enable() and Disable() actions.
** The following options cannot be set at run time:
** - privateColormap
** - mitShm
*/

#define OPT(name) { #name, &appResources.name }
static struct {
  const String name;
  Boolean *value;
} booleanOptions[] = {
  OPT(cache),
  OPT(xsync),
  OPT(paused),
  OPT(emulate1541),
  OPT(exrom),
  OPT(joy1),
  OPT(joy2),
  OPT(confirmExit),
};
#undef OPT

/* X callback functions */

EventHandler(exposure);
EventHandler(structureNotify);

#define ActionProc(name)\
static void name(Widget w, XEvent *event, String *params, Cardinal *num_params)

ActionProc(closeAction);
ActionProc(closeDialogAction);
ActionProc(pasteAction);
ActionProc(toggleAction);
ActionProc(enableAction);
ActionProc(disableAction);

/* Callback for handling paste operations */
static void SelectionReceived (Widget w, XtPointer closure,
			       Atom *selection, Atom *type,
			       XtPointer value,
			       unsigned long *length, int *format);

/*
** Internal functions
*/

/* Install the colormap and allocate the colors */
static Boolean AllocateColors (Widget w);
/* Adjust the window size and other things before realizing it */
static void AdjustWindow (Widget w, Dimension width, Dimension height);
/* Create the image */
static Boolean CreateImage (Dimension width, Dimension height);

/*
** Drawing functions and variables
*/

#if UCHAR_MAX == 255
typedef unsigned char PIXEL1;
#else
#  error "No type for 8-bit values!"
#endif
#if USHRT_MAX == 65535
typedef unsigned short PIXEL2;
#else
#  error "No type for 16-bit values!"
#endif
#if UINT_MAX == 4294967295U
typedef unsigned PIXEL4;
#elif ULONG_MAX == 4294967295U
typedef unsigned long PIXEL4;
#else
#  error "No type for 32-bit values!"
#endif

PIXEL1 pixel1[COLORS];
PIXEL2 pixel2[COLORS];
PIXEL4 pixel4[COLORS];
Pixel pixel[COLORS];

static Display *display = NULL;
static Colormap colormap;

#ifndef NO_SHM
/* Shared memory segment information */
static XShmSegmentInfo shminfo;
#endif /* NO_SHM */
/* Emulator screen image */
static XImage *image = NULL;

void (*convertLine) (unsigned gfx[], unsigned line) = NULL;
static void convertLine1 (unsigned gfx[], unsigned line);
static void convertLine2 (unsigned gfx[], unsigned line);
static void convertLine4 (unsigned gfx[], unsigned line);
static void convertLineX (unsigned gfx[], unsigned line);

/* Emulation geometry */
static struct {
  Dimension width;	/* total image width in pixels */
  Dimension height;	/* total image height in pixels */
  Dimension wwidth;	/* window width in pixels */
  Dimension wheight;	/* window height in pixels */
  Dimension wleft;	/* x offset of the image in the window */
  Dimension wtop;	/* y offset of the image in the window */
} geom;

/* Useful macro for rounding up division result */
#define RoundDiv(a,b) ((a+b-1)/b)

/* flag array: has a vertical section of the image changed? */
static Boolean *cachepartChanged = NULL;
/* number of cache blocks */
static unsigned cacheblocksX = 0;
/* image cache */
static unsigned *gfxcache = NULL;

/* Paste buffer */
static char *pastebuffer = NULL, *newpastebuffer = NULL;

/* Initialize the application. */
Boolean InitializeXt (XtAppContext *appContext, String ClassName,
		      String fallbackRes[],
		      Menu **menus,
		      int width, int height,
		      int *argc, char **argv)
{
  static XtActionsRec actions[] = {
    { "Close", closeAction },
    { "CloseDialog", closeDialogAction },
    { "Paste", pasteAction },
    { "Toggle", toggleAction },
    { "Enable", enableAction },
    { "Disable", disableAction },
  };

  XtSetLanguageProc (NULL, NULL, NULL);

  {
    /* saved command line arguments */
    int xargc; char **xargv;
  
    /* save the command line arguments */
    xargc = *argc;
    xargv = (char**)XtMalloc (xargc * sizeof *xargv);
    memcpy (xargv, argv, xargc * sizeof *xargv);

    /* Create a dummy top level widget in order to determine the visual. */

    if (!(TopLevel = XtAppInitialize (appContext, ClassName,
				      optionList, XtNumber(optionList),
				      argc, argv, fallbackRes, NULL, 0)))
      return False;

    display = XtDisplay (TopLevel);

    XtGetApplicationResources (TopLevel, (XtPointer) &appResources, resources,
			       XtNumber(resources), NULL, 0);

    if (appResources.visual &&
	appResources.visual != DefaultVisualOfScreen (XtScreen (TopLevel))) {
      XVisualInfo vinfo, *vinfo_list;
      Arg args[5];
      int cnt = 0;
      int count;

      XtSetArg (args[cnt], XtNvisual, appResources.visual); cnt++;
      /*
       * Create the colormap.  We could really use the default
       * colormap of the root window for this visual instead.
       */
      colormap = XCreateColormap (display,
				  RootWindowOfScreen (XtScreen (TopLevel)),
				  appResources.visual, AllocNone);
      XtSetArg (args[cnt], XtNcolormap, colormap); cnt++;

      /* Find some information about the visual. */
      vinfo.visualid = XVisualIDFromVisual (appResources.visual);
      vinfo_list = XGetVisualInfo (display, VisualIDMask, &vinfo, &count);

      if (vinfo_list && count > 0) {
	XtSetArg (args[cnt], XtNdepth, vinfo_list[0].depth); cnt++;
	XFree ((XPointer) vinfo_list);
      }

      XtDestroyWidget (TopLevel);

      XtSetArg (args[cnt], XtNargv, xargv); cnt++;
      XtSetArg (args[cnt], XtNargc, xargc); cnt++;

      TopLevel = XtAppCreateShell (NULL, ClassName,
				   applicationShellWidgetClass,
				   display, args, cnt);
    }
    else {
      appResources.visual = DefaultVisualOfScreen (XtScreen (TopLevel));
      appResources.depth = DefaultDepthOfScreen (XtScreen (TopLevel));

      if (!appResources.privateColormap)
	colormap = XDefaultColormapOfScreen (XtScreen (TopLevel));
      else
	colormap = XCreateColormap (display,
				    XRootWindowOfScreen (XtScreen (TopLevel)),
				    appResources.visual, AllocNone);
    }

    XtFree ((XtPointer)xargv);
  }

  /* Initialize the drawing variables */
  geom.width = width;
  geom.height = height;
  geom.wtop = geom.wleft = 0;

  /* Validate the drawing-related resources */
#define CHECK(res,zerocmp,toplimit,default)				\
  do if (appResources.res zerocmp 0 || appResources.res > toplimit) {	\
    fprintf (stderr, #res "=%d out of range, changing to %d\n",		\
	     appResources.res, default);				\
    appResources.res = default;						\
  } while (0)
#define CHECK2(res1,cmp,res2,default1,default2)				\
  do if (appResources.res1 cmp appResources.res2) {			\
    appResources.res1 = default1;					\
    appResources.res2 = default2;					\
    fprintf (stderr,							\
             #res1 #cmp #res2 ", setting " #res1 "=%d, " #res2 "=%d\n",	\
	     appResources.res1, appResources.res2);			\
  } while (0)

  CHECK(speed, <, INT_MAX, 100);

  CHECK(frameParts, <=, geom.height, 1);
  if (geom.height % appResources.frameParts)
    fprintf (stderr,
	     "frameParts not an integer fraction of %d, "
	     "not updating whole frames!", geom.height);

  CHECK(frameSkip, <, INT_MAX, 0);

  CHECK(topLine, <, geom.height - 1, 0);
  CHECK(bottomLine, <=, geom.height, geom.height);
  CHECK(leftColumn, <, geom.width - 1, 0);
  CHECK(rightColumn, <=, geom.width, geom.width);

  CHECK2(topLine, >, bottomLine, 0, geom.height);
  CHECK2(leftColumn, >, rightColumn, 0, geom.width);

  /* Reinitialize the size */
  geom.width = appResources.rightColumn - appResources.leftColumn;
  geom.height = appResources.bottomLine - appResources.topLine;

  CHECK(horizBlock, <=, width, width);
  CHECK(vertBlock, <=, height, height);
  CHECK(fixLeftColumn, <, width - 1, 0);
  CHECK(fixTopLine, <, height - 1, 0);

  CHECK2(fixLeftColumn, <, leftColumn,
	 appResources.leftColumn, appResources.leftColumn);
  CHECK2(fixLeftColumn, >=, rightColumn,
	 appResources.leftColumn, appResources.rightColumn);
  CHECK2(fixTopLine, <, topLine,
	 appResources.topLine, appResources.topLine);
  CHECK2(fixTopLine, >=, bottomLine,
	 appResources.topLine, appResources.bottomLine);
#undef CHECK
#undef CHECK2

  /* Allocate colors */
  if (!AllocateColors (TopLevel)) {
    XtDestroyWidget (TopLevel);
    TopLevel = NULL;
    return False;
  }

  /* Add the action procedures */
  XtAppAddActions(*appContext, actions, XtNumber(actions));

  /* Add the event handlers */
  XtAddEventHandler (TopLevel, ExposureMask, False, exposure, NULL);
  XtAddEventHandler (TopLevel, StructureNotifyMask, False, structureNotify,
		     NULL);

  /* Initialize the cache */
  cacheblocksX = RoundDiv(geom.width, appResources.horizBlock);
  {
    unsigned blocks = cacheblocksX *
      RoundDiv(geom.height, appResources.vertBlock);
    cachepartChanged = calloc (sizeof *cachepartChanged, blocks);
  }

  if (!cachepartChanged || !CreateMenus (TopLevel, menus) ||
      !CreateImage (geom.width, geom.height)) {
    free (cachepartChanged); cachepartChanged = NULL;
    free (gfxcache); gfxcache = NULL;
    XtDestroyWidget (TopLevel);
    TopLevel = NULL;
    return False;
  }

  AdjustWindow (TopLevel, geom.width, geom.height);

  return True;
}

/*
** Static functions
*/

/* Install the colormap and allocate the colors */
static Boolean AllocateColors (Widget w)
{
  int allocated, i;
  XColor screencolor, exactcolor;
  XImage *im;

  /* Map pixels to ZPixmap values. */
  im = XCreateImage (display, appResources.visual,
		     appResources.depth, ZPixmap, 0, NULL, 1, 1, 8, 0);
  if (!im)
    return False;

  if (!(im->data = calloc (im->bytes_per_line, im->height))) {
    XDestroyImage (im);
    return False;
  }

  for (i = allocated = 0; i < COLORS; i++) {
    if (!XAllocNamedColor (display, colormap, appResources.color[i],
			   &screencolor, &exactcolor)) {
      fprintf (stderr, "Warning: cannot allocate color \"%s\"\n",
	       appResources.color[i]);
      /* fallback color */
      screencolor.pixel = i;
    }
    else
      allocated++;

    XPutPixel (im, 0, 0, screencolor.pixel);

    pixel[i] = screencolor.pixel;

    if (8 == im->bits_per_pixel)
      pixel1[i] = *(PIXEL1*)im->data;
    else if (16 == im->bits_per_pixel)
      pixel2[i] = *(PIXEL2*)im->data;
    else if (32 == im->bits_per_pixel)
      pixel4[i] = *(PIXEL4*)im->data;
  }

  if (8 == im->bits_per_pixel)
    convertLine = convertLine1;
  else if (16 == im->bits_per_pixel)
    convertLine = convertLine2;
  else if (32 == im->bits_per_pixel)
    convertLine = convertLine4;
  else {
    convertLine = convertLineX;
    gfxcache = calloc (sizeof *gfxcache, geom.width * geom.height);
  }

  free (im->data);
  im->data = NULL;
  XDestroyImage (im);

  if (allocated < COLORS)
    fprintf (stderr, "Could only get %d of %d colors.\n"
	     "Use the `-install' or '-visual' options "
	     "to install a private colormap.\n",
	     allocated, COLORS);

  /* Set the colormap */
  {
    Arg arg;
    XtSetArg (arg, XtNcolormap, colormap);
    XtSetValues (w, &arg, 1);
  }

  return True;
}

/* Create the image */
static Boolean CreateImage (Dimension width, Dimension height)
{
  /* Check whether the Shared Memory Extension can be used */

#ifndef NO_SHM
  if (appResources.mitShm) {
    char *dpy = DisplayString (display);

    if (*dpy != ':' &&
	strncmp (dpy, "localhost:", 10) &&
	strncmp (dpy, "unix:", 5)) {
      fprintf (stderr, "Warning: not on localhost, MITSHM disabled!\n");
      appResources.mitShm = False;
    }
    else if (!XShmQueryExtension (display)) {
      fprintf (stderr, "Warning: the MITSHM extension is not supported!\n");
      appResources.mitShm = False;
    }
  }

  if (appResources.mitShm) {
    image = XShmCreateImage (display, appResources.visual,
			     appResources.depth,
			     ZPixmap, NULL, &shminfo,
			     width, height);

    shminfo.shmid = shmget (IPC_PRIVATE,
			    image->bytes_per_line * image->height,
			    IPC_CREAT | 0777);

    if (shminfo.shmid == -1) {
      fprintf (stderr, "Cannot get shared memory!\n");
      return False;
    }

    image->data = shminfo.shmaddr = shmat (shminfo.shmid, 0, 0);

    if (!image->data) {
      fprintf (stderr, "Cannot get address for shared memory!\n");
      return False;
    }

    shminfo.readOnly = False;
    if (!XShmAttach (display, &shminfo)) {
      fprintf (stderr, "Cannot attach shared memory.\n");
      return False;
    }
  }
  else {
#else
  appResources.mitShm = False;
  {
#endif /* NO_SHM */
    image = XCreateImage (display, appResources.visual, appResources.depth,
			  ZPixmap, 0, NULL, width, height, 8, 0);

    if (!image) {
      fprintf (stderr, "Could not create the image!\n");
      return False;
    }

    image->data = calloc (image->bytes_per_line, image->height);

    if (!image->data) {
      fprintf (stderr, "Could not allocate memory for the image!\n");
      XDestroyImage (image);
      image = NULL;
      return False;
    }
  }

  return True;
}

/* Adjust the window size and other things before realizing it */
static void AdjustWindow (Widget w, Dimension width, Dimension height)
{
  /* Set some parameters */
  {
    Arg args[5];
    XtSetArg (args[0], XtNwidth, width);
    XtSetArg (args[1], XtNheight, height);
    XtSetArg (args[2], XtNmaxWidth, width);
    XtSetArg (args[3], XtNmaxHeight, height);
    XtSetArg (args[4], XtNinput, True);

    XtSetValues (w, args, XtNumber(args));
  }

  {
    Window window;
    Atom WM_delete_window = XInternAtom (display, "WM_DELETE_WINDOW", False);
    XSetWindowAttributes attr;

    XtRealizeWidget (w);
    window = XtWindow (w);

    attr.background_pixel = attr.border_pixel = pixel[0];
    attr.backing_store =
      !appResources.cache || appResources.mitShm ? NotUseful : Always;

    XChangeWindowAttributes (display, window,
			     CWBackingStore | CWBackPixel | CWBorderPixel,
			     &attr);

    XSetWMProtocols (display, window, &WM_delete_window, 1);
  }
}

/* flag: is the whole cache invalid? */
static Boolean cacheInvalid = True;

/* Convert a line of graphics to an image with PIXEL1 type */
#define ConvertLineFunc(funcname, pixel_t, pixel_d)		\
static void funcname (unsigned gfx[], unsigned line)		\
{								\
  unsigned i;							\
  pixel_t *id;							\
								\
  if (line < appResources.topLine ||				\
      line >= appResources.bottomLine)				\
    return;							\
								\
  line -= appResources.topLine;					\
  gfx += appResources.leftColumn;				\
  id = (pixel_t*)image->data + geom.width * line;		\
								\
  if (cacheInvalid)						\
    for (i = 0; i < geom.width; i++, id++, gfx++)		\
      *id = pixel_d[*gfx & (COLORS - 1)];			\
  else {							\
    unsigned j, k;						\
    Boolean *changed =						\
      &cachepartChanged[(line / appResources.vertBlock) *	\
		       cacheblocksX];				\
								\
    cacheInvalid = True;					\
								\
    for (i = j = 0;						\
	 j < geom.width / appResources.horizBlock + 1;	\
	 j++) {							\
      k = 0;							\
								\
      if (!changed[j]) {					\
	for (; k < appResources.horizBlock && i < geom.width;	\
	     i++, k++, id++, gfx++)				\
	  if (*id != pixel_d[*gfx & (COLORS - 1)]) {		\
	    changed[j] = True;					\
	    break;						\
	  }							\
								\
	if (appResources.cache) cacheInvalid = False;		\
      }								\
								\
      for (; k < appResources.horizBlock && i < geom.width;	\
	   i++, k++, id++, gfx++)				\
	*id = pixel_d[*gfx & (COLORS - 1)];			\
    }								\
  }								\
}

ConvertLineFunc(convertLine1, PIXEL1, pixel1)
ConvertLineFunc(convertLine2, PIXEL2, pixel2)
ConvertLineFunc(convertLine4, PIXEL4, pixel4)

static void convertLineX (unsigned gfx[], unsigned line)
{
  unsigned i;

  if (line < appResources.topLine ||
      line >= appResources.bottomLine)
    return;

  line -= appResources.topLine;
  gfx += appResources.leftColumn;

  if (cacheInvalid || !gfxcache)
    for (i = 0; i < geom.width; i++)
      XPutPixel (image, i, line, pixel[gfx[i] & (COLORS - 1)]);
  else {
    unsigned j, k;
    Boolean *changed =
      &cachepartChanged[(line / appResources.vertBlock) *
		       cacheblocksX];
    unsigned *cache = &gfxcache[line * geom.width];

    cacheInvalid = True;

    for (i = j = 0;
	 j < geom.width / appResources.horizBlock;
	 j++) {
      k = 0;

      if (!changed[j]) {
	for (; k < appResources.horizBlock && i < geom.width;
	     i++, k++, gfx++, cache++)
	  if ((*gfx & (COLORS - 1)) != *cache) {
	    changed[j] = True;
	    break;
	  }

	if (appResources.cache) cacheInvalid = False;
      }

      for (; k < appResources.horizBlock && i < geom.width;
	   i++, k++, gfx++, cache++) {
	XPutPixel (image, i, line, pixel[*gfx & (COLORS - 1)]);
	*cache = *gfx & (COLORS - 1);
      }
    }
  }
}

/* Display a line of graphics (buffered) */
void displayLine (unsigned line)
{
  Window window;
  GC gc;
  unsigned height;
  Boolean *changed;

  if (line < appResources.topLine)
    return;

  line -= appResources.topLine;

  if ((line % appResources.vertBlock) || line >= geom.height)
    return;

  /* UpdateMainWindow () */

  if (line >= geom.wheight + geom.wtop)
    return;

  changed = &cachepartChanged[(line / appResources.vertBlock) * cacheblocksX];
  window = XtWindow (TopLevel);
  gc = XtGetGC (TopLevel, 0, NULL);

  /* Height of the area to be updated */
  height = geom.wheight + geom.wtop - line;
  if (height > appResources.vertBlock)
    height = appResources.vertBlock;

  if (cacheInvalid) {
#ifndef NO_SHM
    if (appResources.mitShm)
      XShmPutImage (display, window, gc, image,
		    geom.wleft, geom.wtop + line, 0, line,
		    geom.width - geom.wleft, height,
		    False);
    else
#endif /* NO_SHM */
      XPutImage (display, window, gc, image,
		 geom.wleft, geom.wtop + line, 0, line,
		 geom.width - geom.wleft, height);
  }
  else {
    unsigned j, k;
    Dimension wline = line;
    if (wline < geom.wtop) {
      height -= geom.wtop - wline;
      wline = geom.wtop;
    }

    for (j = geom.wleft / appResources.horizBlock; j < cacheblocksX; j++) {
      /* Find out how many adjacent blocks have changed */
      for (k = 0; j + k < cacheblocksX && changed[j + k]; k++);

      if (k) {
	Dimension width = geom.width - geom.wleft -
	  j * appResources.horizBlock;

	if (width > k * appResources.horizBlock)
	  width = k * appResources.horizBlock;

#ifndef NO_SHM
	if (appResources.mitShm)
	  XShmPutImage (display, window, gc, image,
			j * appResources.horizBlock, wline,
			j * appResources.horizBlock - geom.wleft,
			wline - geom.wtop,
			width, height,
			False);
	else
#endif /* NO_SHM */
	  XPutImage (display, window, gc, image,
		     j * appResources.horizBlock, wline,
		     j * appResources.horizBlock - geom.wleft,
		     wline - geom.wtop,
		     width, height);

	j += k - 1;
      }
    }
  }

  if (appResources.cache) {
    cacheInvalid = False;
    /* Clear the cache for this line */
    memset (changed, False, sizeof *changed * cacheblocksX);
  }

  if (appResources.xsync)
    XSync (display, False);
}

/* Update the main window */
void UpdateMainWindow ()
{
  structureNotify (TopLevel, NULL, NULL, NULL);
}

/* Perform cleanup */
void Cleanup ()
{
  CloseKeyboard();
  free (cachepartChanged);
  cachepartChanged = NULL;
  free (gfxcache);
  gfxcache = NULL;
  if (newpastebuffer && newpastebuffer != pastebuffer) free (newpastebuffer);
  if (pastebuffer) free (pastebuffer);
  pastebuffer = newpastebuffer = NULL;

  if (display) { /* clean up the display */
    /* to avoid the symptoms of the XShmDetach()/XDestroyImage() bug */
    XSync (display, False);

    if (image) { /* free the image */
#ifndef NO_SHM
      if (shminfo.shmaddr) { /* detach from the shared memory */
	XShmDetach (display, &shminfo);
	XDestroyImage (image);
	shmdt (shminfo.shmaddr);
	shmctl (shminfo.shmid, IPC_RMID, 0);

	shminfo.shmaddr = NULL;
      }
      else {
#else
      {
#endif /* NO_SHM */
	free (image->data);
	image->data = NULL;
	XDestroyImage (image);
      }

      image = NULL;
    }

    display = NULL;
  }

  if (TopLevel) {
    XtAppContext appContext = XtWidgetToApplicationContext (TopLevel);
    XtDestroyWidget (TopLevel);
    TopLevel = NULL;
    XtDestroyApplicationContext (appContext);
  }
}

/* Return any text pasted to the emulator window */
char *GetCommands ()
{
  if (newpastebuffer) {
    if (pastebuffer)
      free (pastebuffer);

    pastebuffer = newpastebuffer;
    newpastebuffer = NULL;

    return pastebuffer;
  }

  return "";
}

EventHandler(structureNotify)
{
  /* Store the old parameters */

  Dimension
    old_wleft = geom.wleft,
    old_wtop = geom.wtop,
    old_wwidth = geom.wwidth,
    old_wheight = geom.wheight;

  if (w != TopLevel) /* We only handle the top level shell */
    return;

  /* Get the window dimensions */
  {
    Arg args[2];
    XtSetArg (args[0], XtNwidth, &geom.wwidth);
    XtSetArg (args[1], XtNheight, &geom.wheight);
    XtGetValues (w, args, XtNumber(args));
  }

  /* Resize the window */
  if (geom.wwidth > geom.width)
    geom.wwidth = geom.width;
  if (geom.wheight > geom.height)
    geom.wheight = geom.height;

  geom.wleft = (geom.width - geom.wwidth) / 2;
  if (appResources.fixLeft &&
      geom.wleft > appResources.fixLeftColumn - appResources.leftColumn)
    geom.wleft = appResources.fixLeftColumn - appResources.leftColumn;

  geom.wtop = (geom.height - geom.wheight) / 2;
  if (appResources.fixTop &&
      geom.wtop > appResources.fixTopLine - appResources.topLine)
    geom.wtop = appResources.fixTopLine - appResources.topLine;

  if (geom.wleft != old_wleft || geom.wtop != old_wtop ||
      geom.width != old_wwidth || geom.height != old_wheight)
    /* Draw any visible changes */
    exposure(w, NULL, NULL, NULL);
}

EventHandler(exposure)
{
  /* We only handle the top level shell. */
  if (w != TopLevel)
    return;

#ifndef NO_SHM
  if (appResources.mitShm)
    XShmPutImage (display, XtWindow (w), XtGetGC (w, 0, NULL), image,
		  geom.wleft, geom.wtop, 0, 0,
		  geom.width - geom.wleft, geom.height - geom.wtop,
		  False);
  else
#endif /* NO_SHM */
    XPutImage (display, XtWindow (w), XtGetGC (w, 0, NULL), image,
	       geom.wleft, geom.wtop, 0, 0,
	       geom.width - geom.wleft, geom.height - geom.wtop);

  if (appResources.xsync)
    XSync (display, False);
}

/* action procedure for closing windows */
ActionProc(closeAction)
{
  static Boolean closing = False;

  if (w != TopLevel) {
    XtUnrealizeWidget (w);
    return;
  }

  if (!closing) {
    closing = True; /* Closing the TopLevel window */

    if (!appResources.confirmExit ||
	Dialog (w, "exit", BT_YES | BT_NO) == btYes)
      closed = True;
    else
      closing = False;
  }
}

/* action procedure for closing dialogs */
ActionProc(closeDialogAction)
{
  /* Find out the parent. */
  while (w && XtParent (w) && XtParent (w) != TopLevel)
    w = XtParent (w);

  if (w && w != TopLevel)
    XtUnrealizeWidget (w);
}

/* paste text to the emulator */
ActionProc(pasteAction)
{
  XtGetSelectionValue (w, XA_PRIMARY, XA_STRING, SelectionReceived,
		       NULL, event->xbutton.time);
}

/* toggle a resource value */
ActionProc(toggleAction)
{
  int i;

  if (!num_params || *num_params != 1)
    return;

  for (i = 0; i < XtNumber(booleanOptions); i++)
    if (!strcmp (booleanOptions[i].name, *params)) {
      *booleanOptions[i].value ^= True;
      UpdateMenus ();
      return;
    }
}

/* turn on a resource value */
ActionProc(enableAction)
{
  int i;

  if (!num_params || *num_params != 1)
    return;

  for (i = 0; i < XtNumber(booleanOptions); i++)
    if (!strcmp (booleanOptions[i].name, *params)) {
      *booleanOptions[i].value = True;
      UpdateMenus ();
      return;
    }
}

/* turn off a resource value */
ActionProc(disableAction)
{
  int i;

  if (!num_params || *num_params != 1)
    return;

  for (i = 0; i < XtNumber(booleanOptions); i++)
    if (!strcmp (booleanOptions[i].name, *params)) {
      *booleanOptions[i].value = False;
      UpdateMenus ();
      return;
    }
}

/* Callback for handling paste operations */
static void SelectionReceived (Widget w, XtPointer closure,
			       Atom *selection, Atom *type,
			       XtPointer value,
			       unsigned long *length, int *format)
{
  int len;

  if (w != TopLevel || !*type || !*length || !value)
    return;

  len = strlen ((char*)value);

  if (!newpastebuffer) {
    strcpy (newpastebuffer = malloc (len + 1), value);
  }
  else {
    newpastebuffer = realloc (newpastebuffer,
			      strlen (newpastebuffer) + len + 1);
    strcat (newpastebuffer, (char*)value);
  }

  XtFree (value);
}
