/*
** Miscellanous utility functions
**
** Copyright  1997 Marko Mkel, see COPYING.LIB for details
*/

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "util.h"

/* Realize and pop up a shell and center it to the parent window */
void Popup (Widget w, Widget parent)
{
  Display *display = XtDisplay (w);

  XtRealizeWidget (w);

  if (parent) {
    Dimension width, height, pwidth, pheight;
    Position x, y, px, py;
    Arg args[2];

    XtVaGetValues (w, XtNwidth, &width, XtNheight, &height, NULL);
    XtVaGetValues (parent, XtNwidth, &pwidth, XtNheight, &pheight,
		   XtNx, &px, XtNy, &py, NULL);
    x = px + (pwidth - width) / 2;
    y = py + (pheight - height) / 2;

    ValidateCoordinates (display, &x, &y, width, height, 0, 0);

    XtSetArg (args[0], XtNx, x);
    XtSetArg (args[1], XtNy, y);
    XtSetValues (w, args, 2);
  }

  {
    Atom WM_delete_window = XInternAtom (display, "WM_DELETE_WINDOW", False);

    XSetWMProtocols(display, XtWindow (w), &WM_delete_window, 1);
  }

  XtPopup (w, XtGrabExclusive);
}

/* Limit the window coordinates to the root window */
void ValidateCoordinates (Display *display, Position *x, Position *y,
			  Dimension width, Dimension height,
			  Dimension xcorr, Dimension ycorr)
{
  int dummy1;
  unsigned int root_width, root_height, dummy2;
  Window root;

  /* Get the root window dimensions */
  XGetGeometry (display, RootWindow (display, XDefaultScreen (display)),
		&root, &dummy1, &dummy1, &root_width, &root_height,
		&dummy2, &dummy2);

  if (*x + width > root_width) { /* Too far to the right */
    if (xcorr > 0)
      *x -= width + xcorr;
    else
      *x = root_width - width;
  }
  if (*y + height > root_height) { /* Too far to the bottom */
    if (ycorr > 0)
      *y -= height + ycorr;
    else
      *y = root_height - height;
  }

  if (*x < 0) *x = 0; /* Fix the left and */
  if (*y < 0) *y = 0; /* top edges as well */
}
